// Copyright 2003 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

import java.net.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import org.bouncycastle.util.encoders.Base64;
import org.xwt.js.*;
import org.xwt.util.*;
import org.xwt.translators.*;

/** Entry point for the XWT Engine; handles splash screen, initial xwar loading, and argument processing */
public class Main {

    /**
     *  FEATURE: this should be implemented using self-emulation
     *  Used for security checks. If this is null, it means that only
     *  scripts originating from the local filesystem are loaded in
     *  the engine (maximum permissions). If scripts have been loaded
     *  from multiple locations, this will be 0.0.0.0 (the invalid
     *  IP).
     */
    public static java.net.InetAddress originAddr = null;
    public static String originHost = null;
    public static String origin = null;

    public static Res builtin = null;

    public static void printUsage() {
        System.err.println("Usage: xwt [-s] [-v] [-l <port>/<url>] source-location [initial-template]");
        System.err.println("");
        System.err.println("Options:");
        // FEATURE: reintroduce
        //System.err.println("    -s show rendering activity with red rectangles");
        System.err.println("    -v verbose logging");
        // FEATURE: reintroduce
        //System.err.println("    -l serve logs via HTTP on 127.0.0.1:<port>/<url>");
        System.err.println("");
        System.err.println("Source-location is one of the following:");
        System.err.println("    - the path to an xwar file");
        System.err.println("    - the path to a directory to be used as an xwar");
        System.err.println("    - the http url of an xwar");
        System.err.println("");
        System.err.println("Initial-template is the resource name of the template to load; defaults to 'main'");
        Runtime.getRuntime().exit(-1);
    }

    /** common entry point */
    public static void main(String[] args) throws Exception {
        int startargs = 0;
        while (true) {
            if (startargs > args.length - 1) printUsage();
            else if (args[startargs].equals("-v")) Log.verbose = true;
            else if (args[startargs].equals("-l")) startargs++;
            else break;
            startargs++;
        }

        Platform.forceLoad();
        if (Log.on) for(int i=0; i<args.length; i++) Log.log(Main.class, "argument " + i + ": " + args[i]);

        builtin = new Res.Zip(new Res.IS(Platform.getBuiltinInputStream()));

        String initialTemplateName = args.length > startargs + 1 ? args[startargs + 1] : "main";
        initialTemplateName = initialTemplateName.replace('/', '.');
        origin = args[startargs];

        if (origin.startsWith("http://") || origin.startsWith("https://")) {
            originHost = origin.substring(origin.indexOf('/') + 2);
            originHost = originHost.substring(0, originHost.indexOf('/') == -1 ? originHost.length() : originHost.indexOf('/'));
            if (originHost.indexOf('@') != -1) originHost = originHost.substring(originHost.indexOf('@') + 1);
            originAddr = InetAddress.getByName(originHost);
        } else {
            // HACK because MSIE turns \'s into /'s in file URLs... argh!!
            if (Platform.platform.getClass().getName().endsWith("Win32")) origin = origin.replace('/', '\\');
            if (!new File(origin).isDirectory()) origin += "!";
            origin = "file:" + origin;
        }

        // FIXME put the splash screen back in
        if (Log.on) Log.log(Main.class, "loading xwar");

        Res rr = Res.stringToRes(origin, true);
        XWT xwt = new XWT(rr);
        Template.getTemplate(((Res)rr.get(initialTemplateName)).addExtension(".xwt")).apply(new Box(), null, xwt);

        Message.Q.startQ();
    }

}
