// Copyright 2003 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;
import java.io.*;
import org.xwt.js.*;
import org.xwt.util.*;
import org.xwt.translators.*;

/** 
 *    <p>
 *    The in-memory representation of a PNG or GIF image. It is
 *    read-only. It is usually passed to PixelBuffer.drawPicture()
 *    </p>
 *
 *    <p>
 *    Implementations of the Platform class should return objects
 *    supporting this interface from the createPicture() method. These
 *    implementations may choose to implement caching strategies (for
 *    example, using a Pixmap on X11).
 *    </p>
 */
public abstract class Picture {

    /** Pictures, cached by Res */
    private static Cache cache = new Cache();

    private static GIF gif = new GIF();
    private static PNG png = new PNG();
    
    /** turns a resource into a Picture.Source */
    public static Picture fromRes(Res r) {
        // FIXME: put self in background thread if needed
        Picture ret = (Picture)cache.get(r);
        if (ret == null) {
            try {
                PushbackInputStream pbis = new PushbackInputStream(r.getInputStream());
                int c = pbis.read();
                pbis.unread(c);
                if (c == 'G') ret = gif.fromInputStream(pbis, r.getDescriptiveName());
                else if (c == 137) ret = png.fromInputStream(pbis, r.getDescriptiveName());
                else if (c == 0xff) ret = Platform.decodeJPEG(pbis, r.getDescriptiveName());
                else throw new JS.Exn("couldn't figure out image type from first byte");
                cache.put(r, ret);
            } catch (IOException e) {
                Log.logJS(Picture.class, e);
                return null;
            }
        }
        return ret;
    }

    /** the height of the picture */
    public abstract int getHeight();
    
    /** the width of the picture */
    public abstract int getWidth();

}
