// Copyright 2003 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

/**
 *  <p>
 *  A block of pixels which can be drawn on and rapidly copied to the
 *  screen.
 *  </p>
 *
 *  <p>
 *  Implementations of the Platform class should return objects
 *  supporting this interface from the _createPixelBuffer()
 *  method. These implementations may choose to use off-screen video
 *  ram for this purpose (for example, a Pixmap on X11).
 *  </p>
 */
public abstract class PixelBuffer {

    /** Draw the region of source within s onto the region d on this PixelBuffer, scaling as needed */
    public abstract void drawPicture(Picture source, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2);

    /** Fill the region (x1, y1, x2, y2) with <tt>color</tt> (AARRGGBB format); the alpha channel component is ignored */
    public abstract void fillRect(int x1, int y1, int x2, int y2, int color);

    /** returns the height of the PixelBuffer */
    public abstract int getHeight();

    /** returns the width of the PixelBuffer */
    public abstract int getWidth();

}
