// Copyright 2003 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt.mips;

import java.util.*;
import java.io.*;

// FIXME: lb/sb/sh/lh need not be word aligned
// FIXME: memory accesses aren't handling sign-extending properly
// FIXME: probably have to implement nonaligned access
// FIXME: implement malloc()

// FIXME: implement an ELF parser based on RandomAccessFile

// FEATURE: progress indicator
// FEATURE: support n32 abi (passes more arguments in registers)
// FEATURE: trap on arithmetic overflows
// FEATURE: FPU
// FEATURE: we always know the value of the pc register; we should emit it as a literal when it appears in computations
// FEATURE: emit bytecode rather than .java code (for on-the-fly classloading without javac present in "real" JVMs)

/** reads a fully linked MIPS ELF binary image on stdin; writes a .java file on stdout */
public class Compiler {

    static String runs = "";
    static int last_emit = -1;
    static DataInputStream dis;
    public static void main(String[] s) throws IOException {

        if (s.length != 2) {
            System.err.println("usage: java " + Compiler.class.getName() + " <classname> <binary.mips>");
            System.exit(-1);
        }

        String packageName = null;
        String className = s[0];
        if (s[0].indexOf('.') != -1) {
            packageName = s[0].substring(0, s[0].lastIndexOf('.'));
            className = s[0].substring(s[0].lastIndexOf('.') + 1);
        }

        System.out.println(prefix + "// This file was generated by MipsToJava");
        if (packageName != null) System.out.println(prefix + "package " + packageName + ";");
        System.out.println(prefix + "public class " + className + " {");
        System.out.println(prefix + "");
        System.out.println(prefix + "    public " + className + "() { }");
        System.out.println(prefix + "");
        System.out.println(prefix + "    // memory");
        System.out.println(prefix + "    public int mem_read[][] = new int[65535][];");
        System.out.println(prefix + "");
        System.out.println(prefix + "    // same as mem_read unless a page is write-protected");
        System.out.println(prefix + "    public int mem_write[][] = new int[65535][];");
        System.out.println(prefix + "");
        System.out.println(prefix + "    // program counter");
        System.out.println(prefix + "    int pc = 0;");
        System.out.println(prefix + "");
        System.out.println(prefix + "    // temporary");
        System.out.println(prefix + "    int tmp = 0;");
        System.out.println(prefix + "");
        System.out.println(prefix + "    // MIPS multiply/divide subsystem; 64-bit result");
        System.out.println(prefix + "    long hilo = 0;");
        System.out.println(prefix + "");
        System.out.println(prefix + "    // General Purpose registers");
        System.out.println(prefix + "    final int r0 = 0;");
        System.out.println(prefix + "    int r1 = 0, r2 = 0, r3 = 0, r4 = 0, r5 = 0, r6 = 0, r7 = 0,");
        System.out.println(prefix + "        r8 = 0, r9 = 0, r10 = 0, r11 = 0, r12 = 0, r13 = 0, r14 = 0, r15 = 0,");
        System.out.println(prefix + "        r16 = 0, r17 = 0, r18 = 0, r19 = 0, r20 = 0, r21 = 0, r22 = 0, r23 = 0,");
        System.out.println(prefix + "        r24 = 0, r25 = 0, r26 = 0, r27 = 0, r28 = 0, r29 = 0, r30 = 0, r31 = 0;");
        System.out.println(prefix + "");

        dis = new DataInputStream(new FileInputStream(s[1]));

        // read the ELF header
        if (dis.readByte() != 0x7f || dis.readByte() != 'E' || dis.readByte() != 'L' || dis.readByte() != 'F')
            throw new RuntimeException("input file is not an ELF binary");
        dis.skip(12);

        if (dis.readShort() != 2)
            throw new RuntimeException("binary is not a linked executable");

        if (dis.readShort() != 8)
            throw new RuntimeException("binary is not a MIPS R3000 binary");

        dis.skip(4);
        
        int entry_point = dis.readInt();
        String entry_point_string = Long.toString(entry_point & 0xffffffffL, 16);

        int ph_offset = dis.readInt();
        int sh_offset = dis.readInt();
        if (ph_offset == 0) throw new RuntimeException("binary is not statically linked");
        dis.skip(4);
        dis.skip(2);

        int ph_entry_size = dis.readShort();
        int ph_num_entries = dis.readShort();
        int sh_entry_size = dis.readShort();
        int sh_num_entries = dis.readShort();
        int string_table_section_number = dis.readShort();

        int skipamount = sh_offset - (4 + 12 + 2 + 2 + 4 + 4 + 4 + 4 + 4 + 2 + 2 + 2 + 2 + 2 + 2);
        while (skipamount>0) skipamount -= (int)dis.skip(skipamount);

        int[] p_type = new int[sh_num_entries];
        int[] p_ofs = new int[sh_num_entries];
        int[] addr = new int[sh_num_entries];
        int[] p_size = new int[sh_num_entries];
        int[] p_name = new int[sh_num_entries];
        for(int i=0; i<sh_num_entries; i++) {
            p_name[i] = dis.readInt();
            p_type[i] = dis.readInt();
            dis.skip(4);
            addr[i] = dis.readInt();
            p_ofs[i] = dis.readInt();
            p_size[i] = dis.readInt();
            dis.skip(sh_entry_size - 4 * 6);
        }

        dis.close();
        dis = new DataInputStream(new FileInputStream(s[1]));

        int seek = p_ofs[string_table_section_number];
        while (seek > 0) seek -= dis.skip(seek);
        char[] stringTable = new char[p_size[string_table_section_number]];
        for(int i=0; i<p_size[string_table_section_number]; i++)
            stringTable[i] = (char)dis.readByte();

        dis.close();
        dis = new DataInputStream(new FileInputStream(s[1]));

        int pos = 0;
        for(int i=0; i<sh_num_entries; i++) {

            String name = "";
            for(int j=p_name[i]; j<stringTable.length && stringTable[j] != 0; j++) name += stringTable[j];
            System.out.println();
            System.out.println(prefix + "// section \"" + name +
                               "\" #" + i + "; file offset 0x" + Long.toString(p_ofs[i] & 0xffffffffL, 16) +
                               ", vma 0x" + Long.toString(addr[i] & 0xffffffffL, 16) + 
                               ", size 0x" + Long.toString(p_size[i] & 0xffffffff, 16));

            if (name.equals(".sdata")) {
        if (last_emit != -1) {
            System.out.println(prefix + "                case 0x" + Long.toString((last_emit & 0xffffffffL) + 0x1000, 16) + ": return;");
            System.out.println(prefix + "                default: throw new Error(\"invalid address 0x\" + Long.toString(pc&0xffffffffL,16));");
            System.out.println(prefix + "        }");
            System.out.println(prefix + "    }");
            last_emit = -1;
        }
                pos = 0; dis.close(); dis = new DataInputStream(new FileInputStream(s[1]));
                while(pos < p_ofs[i]) pos += dis.skip(p_ofs[i] - pos);
                String base = "0x" + Long.toString((addr[i] & 0xffff0000L) >> 16, 16);
                System.out.println(prefix + "    private void initData() {");
                System.out.println(prefix + "        r28 = 0x" + Long.toString((addr[i] - Short.MIN_VALUE - 12) & 0xffffffffL, 16) + ";");
                System.out.println(prefix + "        mem_read[" + base + "] = mem_write[" + base + "] = new int[65535];");
                for(long k=addr[i] & 0xffffffffL; k<((addr[i] + p_size[i]) & 0xffffffffL); k += 4)
                    System.out.println(prefix + "        mem_write[" + base + "][0x" + Long.toString(k & 0xffff, 16) + "] = 0x" +
                                       Long.toString(dis.readInt() & 0xffffffffL, 16) + ";");
                System.out.println(prefix + "    }");

            } else if (name.equals(".text") /*|| name.equals(".init")*/) {
                if (pos > p_ofs[i]) { pos = 0; dis.close(); dis = new DataInputStream(new FileInputStream(s[1])); }
                while(pos < p_ofs[i]) pos += dis.skip(p_ofs[i] - pos);
                int remaining = p_size[i];
                for(int ofs = addr[i]; ofs < addr[i] + p_size[i];) {
                    String base = Long.toString(ofs & 0xffffff00L, 16);
                    int len = Math.min(((ofs + 0x100) & 0xffffff00) - ofs, remaining);
                    emit(ofs, len, dis);
                    last_emit = ofs;
                    remaining -= len;
                    ofs += len;
                }
            }
        }

        if (last_emit != -1) {
            System.out.println(prefix + "                case 0x" + Long.toString((last_emit & 0xffffffffL) + 0x1000, 16) + ": return;");
            System.out.println(prefix + "                default: throw new Error(\"invalid address 0x\" + Long.toString(pc&0xffffffffL,16));");
            System.out.println(prefix + "        }");
            System.out.println(prefix + "    }");
            last_emit = -1;
        }

        System.out.println();
        System.out.println(prefix + "    public static void main(String[] s) { new " + className + "().main(); }");
        System.out.println();
        System.out.println(prefix + "    public void main() {");
        System.out.println();
        System.out.println(prefix + "        // allocate the stack");
        System.out.println(prefix + "        mem_read[1] = mem_write[1] = new int[65535];");
        System.out.println();
        System.out.println(prefix + "        // set the stack pointer");
        System.out.println(prefix + "        r29 = 0x0001ff00;");
        System.out.println();
        System.out.println(prefix + "        // set the \"return address\" from _start to point at the \"magic exit address\" (0xdeadbeef)");
        System.out.println(prefix + "        r31 = 0xdeadbeef;");
        System.out.println();
        System.out.println(prefix + "        // read in the .data segment");
        System.out.println(prefix + "        initData();");
        System.out.println();
        System.out.println(prefix + "        trampoline(0x" + entry_point_string + ");");
        System.out.println();
        System.out.println(prefix + "    }");
        
        System.out.println();
        System.out.println(prefix + "    public void trampoline(int pc) {");
        System.out.println(prefix + "        this.pc = pc;");
        System.out.println(prefix + "        while(true) {");
        System.out.println(prefix + "            switch(this.pc & 0xffffff00) {");
        System.out.println(prefix + "                case 0xdeadbe00: System.out.println(\"exiting with return value \" + r2); System.exit(r2); continue;");
        System.out.print(runs);
        System.out.println(prefix + "                default: throw new Error(\"invalid address 0x\" + Long.toString(this.pc&0xffffffffL,16));");
        System.out.println(prefix + "            }");
        System.out.println(prefix + "        }");
        System.out.println(prefix + "    }");
        System.out.println(prefix + "}");
    }

    static int _instruction;
    static boolean readnext = true;

    /** reads <tt>numbytes</tt> from the stream, emitting <tt>case</tt> blocks starting at vaddr <tt>ofs</tt> */
    static void emit(int vaddr, int numbytes, DataInputStream dis) throws IOException {
        if (last_emit != -1 && ((last_emit & 0xffffff00) != (vaddr & 0xffffff00))) {
            System.out.println(prefix + "                case 0x" + Long.toString((last_emit & 0xffffffffL) + 0x1000, 16) + ": return;");
            System.out.println(prefix + "                default: throw new Error(\"invalid address 0x\" + Long.toString(pc&0xffffffffL,16));");
            System.out.println(prefix + "        }");
            System.out.println(prefix + "    }");
        }
        if (last_emit == -1 || ((last_emit & 0xffffff00) != (vaddr & 0xffffff00))) {
            System.out.println("");
            System.out.println(prefix + "    private void run_" + Long.toString(vaddr & 0xffffff00L,16) + "() {");
            runs += "                case 0x" + Long.toString(vaddr & 0xffffff00L,16) +
                ": run_" + Long.toString(vaddr & 0xffffff00L,16) + "(); continue;\n";
            System.out.println(prefix + "        switch(pc) {");
        }
        int ofs = vaddr;
        try {
            OUTER: for(ofs = vaddr; ofs < vaddr + numbytes; ofs+=4) {
                if (readnext) _instruction = dis.readInt();
                readnext = true;

                String istring = Long.toString(_instruction & 0xffffffffL, 16);
                while(istring.length() < 8) istring = "0" + istring;
                String ostring = Long.toString(ofs & 0xffffffffL, 16);
                while(ostring.length() < 8) ostring = "0" + ostring;
                System.out.print(prefix + "                /* " + istring + " */   case 0x" + ostring + ": System.out.println(\"pc=0x\" + Long.toString(pc&0xffffffffL,16));");

                emit_instruction(ofs, _instruction);
            }
        } catch (EOFException e) {
            emit(ofs, "                // warning, reached EOF before section end");
        } finally {
            last_emit = ofs;
        }  
    }

    private static void emit_instruction(int ofs, int instruction) throws IOException {
        int op = (instruction >>> 26) & 0xff;                // bits 26-31
        int rs = (instruction >> 21) & 0x1f;                 // bits 21-25
        int rt = (instruction >> 16) & 0x1f;                 // bits 16-20 
        int rd = (instruction >> 11) & 0x1f;                 // bits 11-15
        int shamt = (instruction >> 6) & 0x1f;               // bits 6-10
        int subcode = instruction & 0x3f;                    // bits 0-5
        
                int branch_offset = (((instruction & 0x8000) << 16) | ((instruction & 0x7fff))) * 4;
                int jump_target = (instruction & 0x03ffffff) * 4;
                int signed_immediate = (int)((short)(instruction & 0xffff));
                int unsigned_immediate = instruction & 0xffff;
                
                switch(op) {
                    
                    case 0: {
                        switch(subcode) {
                            
                        case 0:
                            if (instruction == 0) emit(ofs, "    /* NOP */;");                     // NOP
                            else emit(ofs, "    r" + rd + " = r" + rt + " << " + shamt + ";");     // SLL
                            break;

                        case 1: throw new Error("opcode 0, subcode 1 is not part of the MIPS I instruction set");

                        case 2:
                            emit(ofs, "    r" + rd + " = r" + rt + " >>> " + shamt + ";");    // SRL
                            break;

                        case 3:
                            emit(ofs, "    r" + rd + " = r" + rt + " >> " + shamt + ";");     // SRA
                            break;

                        case 4:
                            emit(ofs, "    r" + rd + " = r" + rt + " << (r" + rs + " % 32);");     // SLLV
                            break;

                        case 5: throw new Error("opcode 0, subcode 5 is not part of the MIPS I instruction set");

                        case 6:
                            emit(ofs, "    r" + rd + " = r" + rs + " >>> (r" + rt + " % 32);");     // SRLV
                            break;

                        case 7:
                            emit(ofs, "    r" + rd + " = r" + rs + " >> (r" + rt + " % 32);");     // SRAV
                            break;

                        case 8:
                            _instruction = dis.readInt(); readnext = false; emit_instruction(ofs, _instruction);
                            emit(ofs, "    pc = r" + rs + "; return;");                                   // JR
                            break;
                            
                        case 9:
                            _instruction = dis.readInt(); readnext = false; emit_instruction(ofs, _instruction);
                            emit(ofs, "    r" + rd + " = pc + 8; pc = r" + rs + "; return;");             // JALR
                            break;
                            
                        case 10: throw new Error("opcode 0, subcode 10 is not part of the MIPS I instruction set");                            
                        case 11: throw new Error("opcode 0, subcode 11 is not part of the MIPS I instruction set");                            

                        case 12:
                            emit(ofs, "    syscall(" + ((instruction & 0x07ffffc0) >> 6) + ");");         // SYSCALL
                            break;
                            
                        case 13:
                            emit(ofs, "    /* BREAKPOINT */");
                            break;

                        case 14: throw new Error("opcode 0, subcode 14 is not part of the MIPS I instruction set");
                        case 15: throw new Error("opcode 0, subcode 15 is not part of the MIPS I instruction set");

                        case 16:
                            emit(ofs, "    r" + rd + " = (int)(hilo >>> 32);");               // MFHI
                            break;

                        case 17:
                            emit(ofs, "    hilo = (hilo & 0x00000000ffffffffL) | ((r" + rs + " & 0xffffffffL) << 32);");   // MTHI
                            break;
                            
                        case 18:
                            emit(ofs, "    r" + rd + " = (int)hilo;");               // MFLO
                            break;

                        case 19:
                            emit(ofs, "    hilo = (hilo & 0xffffffff00000000L) | (r" + rs + " & 0xffffffffL);");   // MTLO
                            break;

                        case 20: throw new Error("opcode 0, subcode 20 is not part of the MIPS I instruction set");
                        case 22: throw new Error("opcode 0, subcode 22 is not part of the MIPS I instruction set");
                        case 23: throw new Error("opcode 0, subcode 23 is not part of the MIPS I instruction set");
                            
                        case 24:
                            emit(ofs, "    hilo = ((long)r" + rs + ") * ((long)r" + rt + ");");   // MULT
                            break;

                        case 25:
                            emit(ofs, "    hilo = (r" + rs + " & 0xffffffffL) * (r" + rt + " & 0xffffffffL);");   // MULTU
                            break;

                        case 26:
                            emit(ofs, "    hilo = (((r" + rs + " % r" + rt +") & 0xffffffffL) << 32) | ((r" + rs + " / r" + rt +") & 0xffffffffL);");   // DIV
                            break;

                        case 27:
                            emit(ofs, "    hilo = (((r" + rs + " & 0xffffffffL) % (r" + rt +" & 0xffffffffL)) << 32) | " +               // DIVU
                                  "((r" + rs + " & 0xffffffffL) / (r" + rt +" & 0xffffffffL));");   
                            break;

                        case 28: throw new Error("opcode 0, subcode 28 is not part of the MIPS I instruction set");
                        case 29: throw new Error("opcode 0, subcode 29 is not part of the MIPS I instruction set");
                        case 30: throw new Error("opcode 0, subcode 30 is not part of the MIPS I instruction set");
                        case 31: throw new Error("opcode 0, subcode 31 is not part of the MIPS I instruction set");

                        case 32:
                            emit(ofs, "    r" + rd + " = r" + rs + " + r" + rt + ";");                                                    // ADD
                            break;
                            
                        case 33:
                            emit(ofs, "    r" + rd + " = (int)(((r" + rs + " & 0xffffffffL) + (r" + rt + " & 0xffffffffL)));");           // ADDU
                            break;
                            
                        case 34:
                            emit(ofs, "    r" + rd + " = r" + rs + " - r" + rt + ";");                                                    // SUB
                            break;
                            
                        case 35:
                            emit(ofs, "    r" + rd + " = (int)(((r" + rs + " & 0xffffffffL) - (r" + rt + " & 0xffffffffL)));");           // SUBU
                            break;
                            
                        case 36:
                            emit(ofs, "    r" + rd + " = r" + rs + " & r" + rt + ";");                                                    // AND
                            break;

                        case 37:
                            emit(ofs, "    r" + rd + " = r" + rs + " | r" + rt + ";");                                                    // OR
                            break;

                        case 38:
                            emit(ofs, "    r" + rd + " = r" + rs + " ^ r" + rt + ";");                                                    // XOR
                            break;

                        case 39:
                            emit(ofs, "    r" + rd + " = ~(r" + rs + " | r" + rt + ");");                                                 // NOR
                            break;

                        case 40: throw new Error("opcode 0, subcode 40 is not part of the MIPS I instruction set");
                        case 41: throw new Error("opcode 0, subcode 41 is not part of the MIPS I instruction set");

                        case 42:
                            emit(ofs, "    r" + rd + " = (r" + rs + " < r" + rt + ") ? 1 : 0;");                                          // SLT
                            break;

                        case 43:
                            emit(ofs, "    r" + rd + " = ((r" + rs + " & 0xffffffffL) < (r" + rt + " & 0xffffffffL)) ? 1 : 0;");            // SLTU
                            break;

                        case 44: case 45: case 46: case 47: case 48: case 49: case 50: case 51: case 52: case 53: case 54:
                        case 55: case 56: case 57: case 58: case 59: case 60: case 61: case 62: case 63:
                            throw new Error("opcode 0, subcode " + subcode + " is not part of the MIPS I instruction set");                        
                        default:
                            throw new Error("opcode 0, subcode " + subcode + " is not a valid MIPS instruction");
                        }
                        break;
                    }

                case 1: {
                    switch(rt) {
                    case 0:
                            _instruction = dis.readInt(); readnext = false; emit_instruction(ofs, _instruction);
                        emit(ofs, "    if (r" + rs + " < 0) { pc += " + (branch_offset + 4) + "; return; }; ");                            // BLTZ
                        break;

                    case 1:
                            _instruction = dis.readInt(); readnext = false; emit_instruction(ofs, _instruction);
                        emit(ofs, "    if (r" + rs + " >= 0) { pc += " + (branch_offset + 4) + "; return; }; ");                               // BGEZ
                        break;

                    case 16:
                            _instruction = dis.readInt(); readnext = false; emit_instruction(ofs, _instruction);
                        emit(ofs, "    r31 = pc + 4; if (r" + rs + " < 0) { pc += " + (branch_offset + 4) + "; return; }; ");                  // BLTZAL
                        break;

                    case 17:
                            _instruction = dis.readInt(); readnext = false; emit_instruction(ofs, _instruction);
                        emit(ofs, "    r31 = pc + 4; if (r" + rs + " >= 0) { pc += " + (branch_offset + 4) + "; return; }; ");                // BGEZAL
                        break;

                    default: throw new Error("opcode 1, subcode " + rt + " is not part of the MIPS I instruction set");
                    }
                    break;
                }
                    
                case 2:
                            _instruction = dis.readInt(); readnext = false; emit_instruction(ofs, _instruction);
                    emit(ofs, "    pc &= 0xf0000000; pc |= 0x" + Long.toString(jump_target & 0xffffffffL, 16) + "; return;");                                            // J
                    break;

                case 3:
                            _instruction = dis.readInt(); readnext = false; emit_instruction(ofs, _instruction);
                    emit(ofs, "    r31 = pc + 4; pc &= 0xf0000000; pc |= 0x" + Long.toString(jump_target & 0xffffffffL, 16) + "; return;");                                  // JAL
                    break;

                case 4:
                            _instruction = dis.readInt(); readnext = false; emit_instruction(ofs, _instruction);
                    emit(ofs, "    if (r" + rs + " == r" + rt + ") { pc += 0x" + Long.toString((branch_offset + 4) & 0xffffffffL, 16) + "; return; }; ");                  // BEQ
                    break;

                case 5:
                            _instruction = dis.readInt(); readnext = false; emit_instruction(ofs, _instruction);
                    emit(ofs, "    if (r" + rs + " != r" + rt + ") { pc += 0x" + Long.toString((branch_offset + 4) & 0xffffffffL, 16) + "; return; }; ");                  // BNE
                    break;

                case 6:
                            _instruction = dis.readInt(); readnext = false; emit_instruction(ofs, _instruction);
                    emit(ofs, "    if (r" + rs + " <= 0) { pc += 0x" + Long.toString((branch_offset + 4) & 0xffffffffL, 16) + "; return; }; ");                            // BLEZ
                    break;

                case 7:
                            _instruction = dis.readInt(); readnext = false; emit_instruction(ofs, _instruction);
                    emit(ofs, "    if (r" + rs + " > 0) { pc += 0x" + Long.toString((branch_offset + 4) & 0xffffffffL, 16) + "; return; }; ");                             // BGTZ
                    break;

                case 8: case 9:
                    emit(ofs, "    r" + rt + " = r" + rs + " + " + signed_immediate + ";");                                         // ADDI[U]
                    break;
                    
                case 10:
                    emit(ofs, "    r" + rt + " = (r" + rs + " < " + signed_immediate + ") ? 1 : 0;");                               // SLTI
                    break;
                    
                case 11:
                    emit(ofs, "    r" + rt + " = ((r" + rs + " & 0xffffffffL) < (" + signed_immediate + " & 0xffffffffL)) ? 1 : 0;");     // SLTIU
                    break;
                    
                case 12:
                    emit(ofs, "    r" + rt + " = r" + rs + " & " + unsigned_immediate + ";");                                         // ANDI
                    break;
                    
                case 13:
                    emit(ofs, "    r" + rt + " = r" + rs + " | " + unsigned_immediate + ";");                                         // ORI
                    break;
                    
                case 14:
                    emit(ofs, "    r" + rt + " = r" + rs + " ^ " + unsigned_immediate + ";");                                         // XORI
                    break;
                    
                case 15:
                    emit(ofs, "    r" + rt + " = " + unsigned_immediate + " << 16;");                                                 // LUI
                    break;

                case 16: /* throw new Error("coprocessor instructions (opcode 16) are not implemented"); */
		    emit(ofs, "    throw new Error(\"coprocessor instructions (opcode 16) are not implemented\");");
		    break;
                case 17: throw new Error("coprocessor instructions (opcode 17) are not implemented");
                case 18: throw new Error("coprocessor instructions (opcode 18) are not implemented");

                case 19: case 20: case 21: case 22: case 23: case 24: case 25: case 26: case 27: case 28:
                    throw new Error("opcode " + op + " is not part of the MIPS I instruction set @" + Long.toString(ofs & 0xffffffffL, 16));
                    
                case 32: {
                    String dest = "(r" + rs + " + " + signed_immediate + ")";
                    emit(ofs, "    tmp = mem_read[" + dest + ">>>16][" + dest + " & 0xffff];  r" + rt + " = ((tmp & 0x80) << 24) | (tmp & 0x7f);");       // LB
                    break;
                }

                case 33:  {
                    String dest = "(r" + rs + " + " + signed_immediate + ")";
                    emit(ofs, "    tmp = mem_read[" + dest + ">>>16][" + dest + " & 0xffff];  r" + rt + " = ((tmp & 0x8000) << 16) | (tmp & 0x7fff);");   // LH
                    break;
                }

                case 34:
                    emit(ofs, "    throw new Error(\"LWL (opcode 34) is not supported; are you sure you used -mstrict-align?\");");
                    break;

                case 35: {
                    String dest = "(r" + rs + " + " + signed_immediate + ")";
                    emit(ofs, "    r" + rt + " = mem_read[" + dest + ">>>16][" + dest + " & 0xffff];");               // LW
                    break;
                }

                case 36: {
                    String dest = "(r" + rs + " + " + signed_immediate + ")";
                    emit(ofs, "    r" + rt + " = mem_read[" + dest + ">>>16][" + dest + " & 0xffff] & 0xff;");       // LBU
                    break;
                }

                case 37: {
                    String dest = "(r" + rs + " + " + signed_immediate + ")";
                    emit(ofs, "    r" + rt + " = mem_read[" + dest + ">>>16][" + dest + " & 0xffff] & 0xffff;");       // LHU
                    break;
                }

                case 38:
                    emit(ofs, "    throw new Error(\"LWR (opcode 38) is not supported; are you sure you used -mstrict-align?\");");
                    break;

                case 39:  throw new Error("opcode 39 is not part of the MIPS I instruction set");

                case 40: {
                    String dest = "(r" + rs + " + " + signed_immediate + ")";
                    emit(ofs, "    tmp = mem_read[" + dest + ">>>16][" + dest + " & 0xffff]; " +
                          "mem_write[" + dest + ">>>16][" + dest + " & 0xffff] = (tmp & 0xffffff00) | (r" + rt + " & 0xff);");   // SB
                    break;
                }
                    
                case 41: {
                    String dest = "(r" + rs + " + " + signed_immediate + ")";
                    emit(ofs, "    tmp = mem_read[" + dest + ">>>16][" + dest + " & 0xffff]; " +
                          "mem_write[" + dest + ">>>16][" + dest + " & 0xffff] = (tmp & 0xffff0000) | (r" + rt + " & 0xffff);");   // SH
                    break;
                }
                    
                case 42:
                    emit(ofs, "    throw new Error(\"SWL (opcode 42) is not supported; are you sure you used -mstrict-align?\");");
                    break;

                case 43: {
                    String dest = "(r" + rs + " + " + signed_immediate + ")";
                    emit(ofs, "    mem_write[" + dest + ">>>16][" + dest + " & 0xffff] = r" + rt + ";");               // SW
                    break;
                }

                case 44:  throw new Error("opcode 44 is not part of the MIPS I instruction set");
                case 45:  throw new Error("opcode 45 is not part of the MIPS I instruction set");

                case 46:
                    emit(ofs, "    throw new Error(\"SWR (opcode 46) is not supported; are you sure you used -mstrict-align?\");");
                    break;

                case 47:  throw new Error("opcode 47 is not part of the MIPS I instruction set");
                case 48:  throw new Error("opcode 48 is not part of the MIPS I instruction set");

                case 49:
                    emit(ofs, "    throw new Error(\"floating point operations (opcode 49) are not yet supported\");");
                    break;

                case 50:
                    emit(ofs, "    throw new Error(\"floating point operations (opcode 50) are not yet supported\");");
                    break;

                case 51: case 52: case 53: case 54: case 55: case 56:
                    throw new Error("opcode " + op + " is not part of the MIPS I instruction set");

                case 57:
                    emit(ofs, "    throw new Error(\"floating point operations (opcode 57) are not yet supported\");");
                    break;

                case 58:
                    emit(ofs, "    throw new Error(\"floating point operations (opcode 58) are not yet supported\");");
                    break;

                case 60: case 61: case 62: case 63:
                    throw new Error("opcode " + op + " is not part of the MIPS I instruction set");

                default:
                    throw new Error("unknown opcode " + op);
                }
    }


        
    static String prefix = "";
    static void emit(int vaddr, String s) {
        if (s.indexOf("r0 = ") != -1) s = "    /* NOP */";
        if (!s.trim().endsWith("return;") && s.indexOf("throw") == -1) s += " pc = 0x" + Long.toString((vaddr + 4) & 0xffffffffL,16) + ";";
        System.out.println(s);
    }
}

